@@ -111,6 +111,13 @@ struct ldb_val ldb_binary_decode(TALLOC_CTX *mem_ctx, const char *str)
 	return ret;
 }
 
+static bool need_encode(unsigned char cval)
+{
+	if (cval < 0x20 || cval > 0x7E || strchr(" *()\\&|!\"", cval)) {
+		return true;
+	}
+	return false;
+}
 
 /*
    encode a blob as a RFC2254 binary string, escaping any
@@ -124,7 +131,7 @@ char *ldb_binary_encode(TALLOC_CTX *mem_ctx, struct ldb_val val)
 	unsigned char *buf = val.data;
 
 	for (i=0;i<val.length;i++) {
-		if (!isprint(buf[i]) || strchr(" *()\\&|!\"", buf[i])) {
+		if (need_encode(buf[i])) {
 			len += 2;
 		}
 	}
@@ -133,7 +140,7 @@ char *ldb_binary_encode(TALLOC_CTX *mem_ctx, struct ldb_val val)
 
 	len = 0;
 	for (i=0;i<val.length;i++) {
-		if (!isprint(buf[i]) || strchr(" *()\\&|!\"", buf[i])) {
+		if (need_encode(buf[i])) {
 			snprintf(ret+len, 4, "\\%02X", buf[i]);
 			len += 3;
 		} else {
@@ -343,7 +350,7 @@ static enum ldb_parse_op ldb_parse_filtertype(TALLOC_CTX *mem_ctx, char **type,
 	}
 	if (!filter) {
 		talloc_free(name);
-		return filter;
+		return 0;
 	}
 	p++;
 
@@ -535,12 +542,12 @@ static struct ldb_parse_tree *ldb_parse_filterlist(TALLOC_CTX *mem_ctx, const ch
 	while (isspace((unsigned char)*p)) p++;
 
 	while (*p) {
+		struct ldb_parse_tree **e;
 		if (*p == ')') {
 			break;
 		}
 
 		next = ldb_parse_filter(ret->u.list.elements, &p);
-		struct ldb_parse_tree **e;
 		if (next == NULL) {
 			/* an invalid filter element */
 			talloc_free(ret);
@@ -741,7 +748,7 @@ char *ldb_filter_from_tree(TALLOC_CTX *mem_ctx, const struct ldb_parse_tree *tre
 		ret = talloc_asprintf(mem_ctx, "(%s=%s", tree->u.substring.attr,
 				      tree->u.substring.start_with_wildcard?"*":"");
 		if (ret == NULL) return NULL;
-		for (i = 0; tree->u.substring.chunks[i]; i++) {
+		for (i = 0; tree->u.substring.chunks && tree->u.substring.chunks[i]; i++) {
 			s2 = ldb_binary_encode(mem_ctx, *(tree->u.substring.chunks[i]));
 			if (s2 == NULL) {
 				talloc_free(ret);
